#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Reader for bioformat supported files."""

import math
import os
from contextlib import contextmanager
from functools import cached_property, lru_cache
from pathlib import Path
from queue import Empty, SimpleQueue
from tempfile import TemporaryDirectory
from threading import Lock
from typing import Dict, Generator, List, Optional, Type, Union

import jpype.imports  # Needed for loci import to work # noqa
import numpy as np
import ome_types
import scyjava
from jpype.types import JArray
from wsidicom.geometry import Region, Size, SizeMm

"""
Set version of bioformats jar to use with the environmental variable
"BIOFORMATS_VERSION". Note the bioformats jar either has a BSD or GPL-2
license.
"""
bioformats_version = os.getenv("BIOFORMATS_VERSION", "bsd:6.12.0")
scyjava.config.endpoints.append(f"ome:formats-{bioformats_version}")

if not scyjava.jvm_started():
    scyjava.start_jvm()

"""
Load bioformats modules using jpype and scyjava
"""
from loci.common.services import ServiceFactory  # type: ignore # noqa
from loci.formats import ImageReader, Memoizer  # type: ignore # noqa
from loci.formats.services import OMEXMLService  # type: ignore # noqa


class ReaderPool:
    def __init__(
        self,
        filepath: Path,
        max_readers: Optional[int] = None,
        cache_path: Optional[Union[Path, str]] = None,
    ):
        self._filepath = filepath
        if max_readers is None:
            cpu_count = os.cpu_count()
            if cpu_count is not None:
                max_readers = cpu_count
            else:
                max_readers = 1
        self._max_readers = max_readers
        self._current_count = 0
        if cache_path is None:
            self._tempdir = TemporaryDirectory()
            self._cache_path = Path(self._tempdir.name)
        else:
            self._tempdir = None
            self._cache_path = Path(cache_path)
        self._queue: SimpleQueue[Memoizer] = SimpleQueue()
        self._lock = Lock()

    @property
    def filepath(self) -> Path:
        return self._filepath

    @contextmanager
    def get_reader(self) -> Generator[Memoizer, None, None]:
        """Return a reader. Should be used as a context manager. Will block
        if no reader is available."""
        reader = self._get_reader()
        try:
            yield reader
        finally:
            self._return_reader(reader)

    @property
    def _is_full(self) -> bool:
        return self._current_count == self._max_readers

    @property
    def _is_empty(self) -> bool:
        return self._current_count == 0

    def _increment(self):
        if self._current_count < self._max_readers:
            self._current_count += 1
            return
        raise ValueError("Current count should never be higher than max readers.")

    def _decrement(self):
        if self._current_count > 0:
            self._current_count -= 1
            return
        raise ValueError("Current count should never be lower than 0.")

    def _get_reader(self) -> Memoizer:
        """Return a reader with no wait if one is available. Else if current
        reader count is less than maximum reader count return a new reader.
        Otherwise wait for an available readern."""
        try:
            return self._queue.get_nowait()
        except Empty:
            pass
        with self._lock:
            if not self._is_full:
                self._increment()
                return self._create_new_reader()

        return self._queue.get(block=True, timeout=None)

    def _create_new_reader(self) -> Memoizer:
        """Return a new reader."""
        # Create a reader using Memoizer to load file faster
        # See https://docs.openmicroscopy.org/bio-formats/6.11.0/developers/matlab-dev.html#reader-performance
        reader = Memoizer(ImageReader(), 0, self._cache_path)
        reader.setFlattenedResolutions(False)
        reader.setId(str(self._filepath))
        return reader

    def _return_reader(self, reader: Memoizer) -> None:
        """Release a used reader."""
        self._queue.put(reader)

    def close(self) -> None:
        """Close all readers and clean up cache directory if a temporary directory."""
        with self._lock:
            while not self._is_empty:
                reader = self._queue.get()
                reader.close()
                self._decrement()
            if self._tempdir is not None:
                self._tempdir.cleanup()


class BioformatsReader:
    def __init__(
        self,
        filepath: Path,
        max_readers: Optional[int] = None,
        cache_path: Optional[Union[Path, str]] = None,
    ):
        """Reader for image data and metadata from file using Bio-Formats api.

        Parameters
        ----------
        filepath: Path
            Path to file to open.
        max_readers: int
            Maximum number of readers to use. If not specified the number
            of cpus will be used.
        cache_path: Optional[Union[Path, str]] = None
            Path to store cache file quicker opening of new readers. If None
            a temporary directory will be used.
        """
        self._reader_pool = ReaderPool(filepath, max_readers, cache_path)

    @staticmethod
    def is_supported(filepath: Path) -> bool:
        try:
            reader = ImageReader()
            reader.setId(str(filepath))
            reader.close()
        except Exception:
            return False
        return True

    @property
    def filepath(self) -> Path:
        """Return filepath of the opened file."""
        return self._reader_pool.filepath

    @cached_property
    def metadata(self) -> ome_types.OME:
        """Return parsed metadata."""
        metadata = self._read_metadata()
        return ome_types.from_xml(str(metadata), parser="lxml")

    @property
    def images_count(self) -> int:
        """Return number of images in file."""
        return len(self.metadata.images)

    def image_name(self, image_index: int) -> Optional[str]:
        """Return name of image."""
        return self.metadata.images[image_index].name

    @lru_cache
    def dtype(self, image_index: int) -> np.dtype:
        """Return the numpy datatype for image in file."""
        NUMPY_DATA_TYPES: Dict[str, Type] = {
            "bit": np.bool_,
            "double-complex": np.cdouble,
            "complex": np.csingle,
            "double": np.float64,
            "float": np.float32,
            "int16": np.int16,
            "int32": np.int32,
            "int8": np.int8,
            "uint16": np.uint16,
            "uint32": np.uint32,
            "uint8": np.uint8,
        }
        if self.metadata.images[image_index].pixels.big_endian:
            byte_order = ">"
        else:
            byte_order = "<"
        data_type = self.metadata.images[image_index].pixels.type.value
        try:
            numpy_data_type = np.dtype(NUMPY_DATA_TYPES[data_type])
        except KeyError:
            raise ValueError(f"Unkown data type {data_type}")
        return numpy_data_type.newbyteorder(byte_order)

    @lru_cache
    def samples_per_pixel(self, image_index: int) -> int:
        """Return the samples per pixel for image in file."""
        pixels = self.metadata.images[image_index].pixels
        samples_per_pixel = pixels.channels[0].samples_per_pixel
        assert samples_per_pixel is not None
        return int(samples_per_pixel)

    @lru_cache
    def size(self, image_index: int, resolution_index: int = 0) -> Size:
        """Return the image size for image in file."""
        with self._reader_pool.get_reader() as reader:
            reader.setSeries(image_index)
            reader.setResolution(resolution_index)
            width = reader.getSizeX()
            height = reader.getSizeY()
            return Size(int(width), int(height))

    @lru_cache
    def pixel_spacing(
        self, image_index: int, resolution_index: int = 0
    ) -> Optional[SizeMm]:
        """Return the size of the pixels in mm/pixel for image in file."""
        pixels = self.metadata.images[image_index].pixels
        if pixels.physical_size_x is None or pixels.physical_size_y is None:
            return None
        scale = self._resolution_scales(image_index)[resolution_index]
        return (
            SizeMm(float(pixels.physical_size_x), float(pixels.physical_size_y))
            * scale
            / 1000
        )

    @lru_cache
    def is_interleaved(self, image_index: int) -> bool:
        """Return true if image data is interleaved."""
        interleaved = self.metadata.images[image_index].pixels.interleaved
        assert interleaved is not None
        return interleaved

    @lru_cache
    def pyramid_levels(self, image_index: int) -> Dict[int, int]:
        """Return dictionary of dyadic scaling as key and resolution index as value
        for resolutions in image."""
        TOLERANCE = 1e-2
        float_pyramid_levels = (
            math.log2(scale) for scale in self._resolution_scales(image_index)
        )
        return {
            round(float_level): resolution_index
            for resolution_index, float_level in enumerate(float_pyramid_levels)
            if math.isclose(float_level, round(float_level), abs_tol=TOLERANCE)
        }

    @lru_cache
    def _resolution_scales(self, image_index: int) -> List[float]:
        """Return resolution scales for image as resolution width divided by image width."""
        with self._reader_pool.get_reader() as reader:
            reader.setSeries(image_index)
            reader.setResolution(0)
            base_width = reader.getSizeX()
            return [
                self._get_resolution_scale(reader, resolution_index, base_width)
                for resolution_index in range(reader.getResolutionCount())
            ]

    @staticmethod
    def _get_resolution_scale(
        reader: Memoizer, resolution_index: int, base_width: int
    ) -> float:
        """Return resolution scale for resolution as rounded int of resolution width
        divided by base widht."""
        reader.setResolution(resolution_index)
        width = reader.getSizeX()
        return base_width / width

    @contextmanager
    def read_image(
        self,
        image_index: int,
        resolution_index: int,
        region: Region,
        output_size: Optional[Size] = None,
        index: int = 0,
    ) -> Generator[np.ndarray, None, None]:
        """Read image data from file. Preferably used as a context manager.
        Data is returned as a memoryview that should be released after use.

        Parameters
        ----------
        image_index: int
            The image to read data from.
        region: Region
            The region to read data from.
        output_size: Optional[Size] = None
            Optional size to resize image data to. Must be equal to or larger than
            region size.
        index: int
            The index in image to read data from.

        Returns
        ----------
            Generator[np.ndarray, None, None]

        """
        try:
            raw_data = memoryview(
                self._read(
                    image_index,
                    resolution_index,
                    index,
                    region.start.x,
                    region.start.y,
                    region.size.width,
                    region.size.height,
                )
            )
        except Exception as exception:
            raise Exception(
                f"Failed to read image data from image {image_index}, resolution {resolution_index}, index {index} at region {region.box}."
            ) from exception
        try:
            data: np.ndarray = np.frombuffer(raw_data, self.dtype(image_index))
            if self.is_interleaved(image_index):
                data = data.reshape(
                    region.size.height,
                    region.size.width,
                    self.samples_per_pixel(image_index),
                )
            else:
                data = data.reshape(
                    self.samples_per_pixel(image_index),
                    region.size.height,
                    region.size.width,
                )
                data = np.moveaxis(data, 0, 2).copy()
            if output_size is not None and data.shape[0:2] != output_size.to_tuple():
                # Pad with zeros to get requsted output size.
                if not output_size.all_greater_than_or_equal(region.size):
                    raise ValueError(
                        "Output size should be equal to or larger than region size."
                    )
                padding_width = output_size.width - data.shape[0]
                padding_height = output_size.height - data.shape[1]
                data = np.pad(data, ((0, padding_width), (0, padding_height), (0, 0)))
            yield data
        finally:
            raw_data.release()

    def close(self) -> None:
        """Close reader pool."""
        self._reader_pool.close()

    def _read(
        self,
        image_index: int,
        resolution_index: int,
        index: int,
        start_x: int,
        start_y: int,
        end_x: int,
        end_y: int,
    ) -> JArray:
        """Read image data from file."""
        with self._reader_pool.get_reader() as reader:
            reader.setSeries(image_index)
            reader.setResolution(resolution_index)
            return reader.openBytes(index, start_x, start_y, end_x, end_y)

    def _read_metadata(self) -> str:
        """Read metadata from file."""
        with self._reader_pool.get_reader() as reader:
            metadata_store = reader.getMetadataStore()
            return str(metadata_store.dumpXML())



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Base model for metadata."""
from dataclasses import Field, fields, is_dataclass, replace
from typing import Any, Optional, Sequence, Type, TypeVar

from PIL import ImageCms
from pydicom.uid import UID
from wsidicom.metadata import (
    Equipment,
    Image,
    Label,
    OpticalPath,
    Patient,
    Series,
    Slide,
    Study,
    WsiMetadata,
)

ModelType = TypeVar("ModelType")


class WsiDicomizerMetadata(WsiMetadata):
    """Subclass of WsiMetadata that can be constructed with optional parameters."""

    def __init__(
        self,
        study: Optional[Study] = None,
        series: Optional[Series] = None,
        patient: Optional[Patient] = None,
        equipment: Optional[Equipment] = None,
        optical_paths: Optional[Sequence[OpticalPath]] = None,
        slide: Optional[Slide] = None,
        label: Optional[Label] = None,
        image: Optional[Image] = None,
        frame_of_reference_uid: Optional[UID] = None,
        dimension_organization_uids: Optional[Sequence[UID]] = None,
    ):
        super().__init__(
            study if study is not None else Study(),
            series if series is not None else Series(),
            patient if patient is not None else Patient(),
            equipment if equipment is not None else Equipment(),
            optical_paths if optical_paths is not None else [],
            slide if slide is not None else Slide(),
            label if label is not None else Label(),
            image if image is not None else Image(),
            frame_of_reference_uid,
            dimension_organization_uids,
        )

    def merge(
        self,
        user: Optional[WsiMetadata],
        default: Optional[WsiMetadata],
        include_confidential: bool,
    ) -> "WsiDicomizerMetadata":
        if not include_confidential:
            base = self._remove_confidential()
        else:
            base = self
        if user is None and default is None:
            return self._merge_not_none(WsiDicomizerMetadata, base, None, None)
        if user is None:
            user = WsiDicomizerMetadata()
        if default is None:
            default = WsiDicomizerMetadata()
        return WsiDicomizerMetadata(
            study=self._merge(Study, base.study, user.study, default.study),
            series=self._merge(Series, base.series, user.series, default.series),
            patient=self._merge(Patient, base.patient, user.patient, default.patient),
            equipment=self._merge(
                Equipment, base.equipment, user.equipment, default.equipment
            ),
            optical_paths=self._merge_list(
                OpticalPath,
                base.optical_paths,
                user.optical_paths,
                default.optical_paths,
            ),
            slide=self._merge(Slide, base.slide, user.slide, default.slide),
            label=self._merge(Label, base.label, user.label, default.label),
            image=self._merge(Image, base.image, user.image, default.image),
        )

    def _remove_confidential(self) -> "WsiDicomizerMetadata":
        return WsiDicomizerMetadata(
            study=None,
            series=None,
            patient=self.patient.remove_confidential() if self.patient else None,
            equipment=self.equipment.remove_confidential() if self.equipment else None,
            optical_paths=self.optical_paths,
            slide=None,
            label=self.label.remove_confidential() if self.label else None,
            image=self.image.remove_confidential() if self.image else None,
            frame_of_reference_uid=None,
            dimension_organization_uids=None,
        )

    @classmethod
    def _merge_list(
        cls,
        model_class: Type[ModelType],
        base: Sequence[ModelType],
        user: Sequence[ModelType],
        default: Sequence[ModelType],
    ) -> Sequence[ModelType]:
        models = [model for model in (user, base, default) if len(model) > 0]
        if len(models) == 0 or all(item is None for model in models for item in model):
            # All lists empty
            return []
        if len(models) == 1:
            # Only one list not empty
            return models[0]
        user_expanded = cls._repeat_list(base, user)
        default_expanded = cls._repeat_list(base, default)
        return [
            cls._merge_not_none(model_class, base_item, user_item, default_item)
            for base_item, user_item, default_item in zip(
                base, user_expanded, default_expanded
            )
        ]

    @staticmethod
    def _repeat_list(
        base: Sequence[ModelType], to_repeat: Sequence[ModelType]
    ) -> Sequence[Optional[ModelType]]:
        if len(to_repeat) > 1 and len(to_repeat) != len(base):
            raise ValueError(
                "List to repeat must have length 0, 1 or length of base. "
                f"Length of list to repeat: {len(to_repeat)}. "
                f"Length of base: {len(base)}."
            )
        if len(to_repeat) == len(base):
            return to_repeat
        if len(to_repeat) == 0:
            item_to_repeat = None
        else:
            item_to_repeat = to_repeat[0]
        return [item_to_repeat for _ in range(len(base))]

    @classmethod
    def _merge(
        cls,
        model_class: Type[ModelType],
        base: Optional[ModelType],
        user: Optional[ModelType],
        default: Optional[ModelType],
    ) -> Optional[ModelType]:
        """Merge three models to a new model.

        - base - model from file.
        - user - user defined model that overrides base.
        - default - user defined model with default values used for fields not defined
        in base or user model.

        """
        not_none = [model for model in [user, base, default] if model is not None]
        if len(not_none) == 0:
            return None
        assert is_dataclass(model_class)
        attributes = {
            field.name: cls._select_value(field, base, user, default)
            for field in fields(model_class)
        }

        return model_class(**attributes)

    @classmethod
    def _merge_not_none(
        cls,
        model_class: Type[ModelType],
        base: Optional[ModelType],
        user: Optional[ModelType],
        default: Optional[ModelType],
    ) -> ModelType:
        merged = cls._merge(model_class, base, user, default)
        assert merged is not None
        return merged

    @classmethod
    def _select_value(
        cls,
        field: Field,
        base: Optional[ModelType],
        user: Optional[ModelType],
        default: Optional[ModelType],
    ) -> Any:
        base_value = getattr(base, field.name, None)
        user_value = getattr(user, field.name, None)
        default_value = getattr(default, field.name, None)
        value = next(
            (
                value
                for value in [user_value, base_value, default_value]
                if value is not None
            ),
            None,
        )
        if is_dataclass(value) and isinstance(value, object):
            value = cls._merge(value.__class__, base_value, user_value, default_value)
        return value

    def insert_default_icc_profile(self) -> WsiMetadata:
        if len(self.optical_paths) == 0:
            # No optical paths defined, add one with icc profile
            optical_paths = [
                OpticalPath(icc_profile=self._create_default_icc_profile())
            ]
        else:
            # Optical paths defined, add icc profile if missing
            optical_paths = [
                (
                    replace(
                        optical_path,
                        icc_profile=self._create_default_icc_profile(),
                    )
                    if optical_path.icc_profile is None
                    else optical_path
                )
                for optical_path in self.optical_paths
            ]
        return replace(
            self,
            optical_paths=optical_paths,
            frame_of_reference_uid=self.default_frame_of_reference_uid,
            dimension_organization_uids=self.default_dimension_organization_uids,
        )

    @staticmethod
    def _create_default_icc_profile() -> bytes:
        return ImageCms.ImageCmsProfile(ImageCms.createProfile("sRGB")).tobytes()



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module containing a base Source implementation suitable for use with non-DICOM
files."""

from abc import ABCMeta, abstractmethod
from functools import cached_property
from pathlib import Path
from typing import Dict, List, Optional, Sequence

from pydicom import config
from wsidicom import ImageData
from wsidicom.codec import Encoder
from wsidicom.graphical_annotations import AnnotationInstance
from wsidicom.instance import ImageType, WsiDataset, WsiInstance
from wsidicom.metadata import WsiMetadata
from wsidicom.metadata.schema.dicom import WsiMetadataDicomSchema
from wsidicom.source import Source

from wsidicomizer.config import settings
from wsidicomizer.image_data import DicomizerImageData
from wsidicomizer.metadata import WsiDicomizerMetadata

config.enforce_valid_values = True
config.future_behavior()


class DicomizerSource(Source, metaclass=ABCMeta):
    """
    Metaclass for Dicomizer sources. Subclasses should implement the method
    is_supported(), _create_level_image_data(), _create_label_image_data(), and
     _create_overview_image_data() and the properties metadata, pyramid_levels.
     Subclasses can override the __init__().
    """

    def __init__(
        self,
        filepath: Path,
        encoder: Encoder,
        tile_size: Optional[int] = None,
        metadata: Optional[WsiMetadata] = None,
        default_metadata: Optional[WsiMetadata] = None,
        include_confidential: bool = True,
    ) -> None:
        self._filepath = filepath
        self._encoder = encoder
        self._tile_size = tile_size
        self._user_metadata = metadata
        self._default_metadata = default_metadata
        self._include_confidential = include_confidential

    @staticmethod
    @abstractmethod
    def is_supported(path: Path) -> bool:
        """Return True if file in filepath is supported by Dicomizer."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def base_metadata(self) -> WsiDicomizerMetadata:
        """Return metadata for file."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def pyramid_levels(self) -> Dict[int, int]:
        """Dictionary of pyramid levels (scalings) as key and level index in file as
        value for levels in file."""
        raise NotImplementedError()

    @abstractmethod
    def _create_level_image_data(self, level_index: int) -> DicomizerImageData:
        """Return image data instance for level."""
        raise NotImplementedError()

    @abstractmethod
    def _create_label_image_data(self) -> Optional[DicomizerImageData]:
        """Return image data instance for label."""
        raise NotImplementedError()

    @abstractmethod
    def _create_overview_image_data(self) -> Optional[DicomizerImageData]:
        """Return image data instance for overview."""
        raise NotImplementedError()

    @abstractmethod
    def _create_thumbnail_image_data(self) -> Optional[DicomizerImageData]:
        """Return image data instance for thumbnail."""
        raise NotImplementedError()

    @cached_property
    def metadata(self) -> WsiDicomizerMetadata:
        return self.base_metadata.merge(
            self.user_metadata, self.default_metadata, self._include_confidential
        )

    @property
    def user_metadata(self) -> Optional[WsiMetadata]:
        return self._user_metadata

    @property
    def default_metadata(self) -> Optional[WsiMetadata]:
        return self._default_metadata

    @property
    def base_dataset(self) -> WsiDataset:
        return self.level_instances[0].dataset

    @cached_property
    def level_instances(self) -> List[WsiInstance]:
        return [
            self._create_instance(
                self._create_level_image_data(level_index),
                ImageType.VOLUME,
                pyramid_index,
            )
            for pyramid_index, level_index in self.pyramid_levels.items()
        ]

    @cached_property
    def label_instances(self) -> List[WsiInstance]:
        label = self._create_label_image_data()
        if label is None:
            return []
        return [self._create_instance(label, ImageType.LABEL)]

    @cached_property
    def overview_instances(self) -> List[WsiInstance]:
        overview = self._create_overview_image_data()
        if overview is None:
            return []
        return [self._create_instance(overview, ImageType.OVERVIEW)]

    @cached_property
    def thumbnail_instances(self) -> List[WsiInstance]:
        thumbnail = self._create_thumbnail_image_data()
        if thumbnail is None:
            return []
        return [self._create_instance(thumbnail, ImageType.THUMBNAIL)]

    @property
    def annotation_instances(self) -> List[AnnotationInstance]:
        return []

    def _create_instance(
        self,
        image_data: ImageData,
        image_type: ImageType,
        pyramid_index: Optional[int] = None,
    ) -> WsiInstance:
        """Create instance from image data."""
        dataset = self._create_dataset(
            image_type, image_data.photometric_interpretation
        )
        return WsiInstance.create_instance(
            image_data, dataset, image_type, pyramid_index
        )

    def _create_dataset(
        self, image_type: ImageType, photometric_interpretation: str
    ) -> WsiDataset:
        if (
            settings.insert_icc_profile_if_missing
            and not photometric_interpretation.startswith("MONOCHROME")
        ):
            metadata = self.metadata.insert_default_icc_profile()
        else:
            metadata = self.metadata
        dataset = WsiMetadataDicomSchema(context={"image_type": image_type}).dump(
            metadata
        )
        return WsiDataset(dataset)

    @staticmethod
    def _is_included_level(
        level: int,
        present_levels: Sequence[int],
        include_indices: Optional[Sequence[int]] = None,
    ) -> bool:
        """Return true if pyramid level is in included levels.

        Parameters
        ----------
        level: int
            Pyramid level to check.
        present_levels: Sequence[int]
            List of pyramid levels present.
        include_indices: Optional[Sequence[int]] = None
            Optional list indices (in present levels) to include, e.g. [0, 1]
            includes the two lowest levels. Negative indices can be used,
            e.g. [-1, -2] includes the two highest levels. Default of None
            will not limit the selection. An empty sequence will excluded all
            levels.

        Returns
        ----------
        bool
            True if level should be included.
        """
        if level not in present_levels:
            return False
        if include_indices is None:
            return True
        absolute_levels = [
            present_levels[level]
            for level in include_indices
            if -len(present_levels) <= level < len(present_levels)
        ]
        return level in absolute_levels



#    Copyright 2021 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import argparse
import os
from pathlib import Path
from typing import Optional, Sequence

from wsidicom.codec import Jpeg2kSettings, JpegSettings, Settings, Subsampling
from wsidicom.metadata.schema.json.wsi import WsiMetadataJsonSchema
from wsidicom.metadata.wsi import WsiMetadata

from wsidicomizer.wsidicomizer import WsiDicomizer


class WsiDicomizerCli:
    def __init__(self):
        self._parser = argparse.ArgumentParser(
            description=("Convert compatible wsi file to DICOM")
        )

        self._parser.add_argument(
            "-i", "--input", type=Path, required=True, help="Path to input wsi file."
        )
        self._parser.add_argument(
            "-o",
            "--output",
            type=Path,
            help=(
                "Path to output folder. Folder will be created and must not "
                "exist. If not specified a folder named after the input file "
                "is created in the same path."
            ),
        )
        self._parser.add_argument(
            "-t",
            "--tile-size",
            type=int,
            default=512,
            help=(
                "Tile size (same for width and height). Required for ndpi and "
                "openslide formats E.g. 512"
            ),
        )
        self._parser.add_argument(
            "-m",
            "--metadata",
            type=Path,
            help=(
                "Path to json metadata that will override metadata from source image "
                "file."
            ),
        )
        self._parser.add_argument(
            "-d",
            "--default-metadata",
            type=Path,
            help=("Path to json metadata that will be used as default values."),
        )
        self._parser.add_argument(
            "-l",
            "--levels",
            type=int,
            nargs="+",
            help=(
                "Pyramid levels to include, if not all. E.g. 0 1 for base and "
                "first pyramid layer. "
            ),
        )
        self._parser.add_argument(
            "--add-missing-levels",
            action="store_true",
            help="If to add missing dyadic levels up to the single tile level.",
        )
        self._parser.add_argument(
            "--label",
            type=Path,
            help="Optional label image to use instead of label found in file.",
        )
        self._parser.add_argument(
            "--no-label", action="store_true", help="If not to include label"
        )
        self._parser.add_argument(
            "--no-overview", action="store_true", help="If not to include overview"
        )
        self._parser.add_argument(
            "--no-confidential",
            action="store_true",
            help="If not to include confidential metadata",
        )
        self._parser.add_argument(
            "-w",
            "--workers",
            type=int,
            default=os.cpu_count(),
            help="Number of worker threads to use",
        )
        self._parser.add_argument(
            "--chunk-size",
            type=int,
            default=100,
            help="Number of tiles to give each worker at a time",
        )
        self._parser.add_argument(
            "--format",
            type=str,
            default="jpeg",
            help="Encoding format to use if re-encoding. 'jpeg' or 'jpeg2000'.",
        )
        self._parser.add_argument(
            "--quality",
            type=float,
            default=90,
            help=(
                "Quality to use if re-encoding. It is recommended to not use > 95 for "
                "jpeg. Use < 1 or > 1000 for lossless jpeg2000."
            ),
        )
        self._parser.add_argument(
            "--subsampling",
            type=str,
            default="420",
            help=(
                "Subsampling option if using jpeg for re-encoding. Use '444' "
                "for no subsampling, '422' for 2x1 subsampling, and '420' for "
                "2x2 subsampling."
            ),
        )
        self._parser.add_argument(
            "--offset-table",
            type=str,
            default="bot",
            help=(
                "Offset table to use, 'bot' basic offset table, 'eot' "
                "extended offset table, 'None' - no offset table."
            ),
        )

    def cli(self):
        args = self._parser.parse_args()
        if not args.metadata:
            metadata = None
        else:
            metadata = self._load_metadata(args.metadata)
        if not args.default_metadata:
            default_metadata = None
        else:
            default_metadata = self._load_metadata(args.default_metadata)
        if not args.levels:
            levels = None
        else:
            levels = args.levels
        encoding_format = args.format
        if encoding_format == "jpeg":
            subsampling = Subsampling.from_string(args.subsampling)
            encoding_settings = JpegSettings(
                quality=args.quality, subsampling=subsampling
            )
        elif encoding_format == "jpeg2000":
            encoding_settings = Jpeg2kSettings(levels=[args.quality])
        else:
            encoding_settings = None
        self.convert(
            filepath=args.input,
            output_path=args.output,
            metadata=metadata,
            default_metadata=default_metadata,
            tile_size=args.tile_size,
            add_missing_levels=args.add_missing_levels,
            include_levels=levels,
            include_label=not args.no_label,
            include_overview=not args.no_overview,
            include_confidential=not args.no_confidential,
            workers=args.workers,
            chunk_size=args.chunk_size,
            encoding_settings=encoding_settings,
            offset_table=args.offset_table,
            label=args.label,
        )

    def convert(
        self,
        filepath: Path,
        output_path: Path,
        metadata: Optional[WsiMetadata] = None,
        default_metadata: Optional[WsiMetadata] = None,
        tile_size: int = 512,
        add_missing_levels: bool = False,
        include_levels: Optional[Sequence[int]] = None,
        include_label: bool = True,
        include_overview: bool = True,
        include_confidential: bool = True,
        workers: Optional[int] = None,
        chunk_size: Optional[int] = None,
        encoding_settings: Optional[Settings] = None,
        offset_table: str = "bot",
        label: Optional[Path] = None,
    ):
        WsiDicomizer.convert(
            filepath=filepath,
            output_path=output_path,
            metadata=metadata,
            default_metadata=default_metadata,
            tile_size=tile_size,
            add_missing_levels=add_missing_levels,
            include_levels=include_levels,
            include_label=include_label,
            include_overview=include_overview,
            include_confidential=include_confidential,
            workers=workers,
            chunk_size=chunk_size,
            encoding=encoding_settings,
            offset_table=offset_table,
            label=label,
        )

    @staticmethod
    def _load_metadata(filepath: Path) -> WsiMetadata:
        with open(filepath) as json_file:
            metadata = WsiMetadataJsonSchema().loads(json_file.read())
            assert isinstance(metadata, WsiMetadata)
            return metadata


def main():
    cli = WsiDicomizerCli()
    cli.cli()


if __name__ == "__main__":
    main()



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module containing a base ImageData implementation suitable for use with non-DICOM
files."""

from abc import ABCMeta, abstractmethod
from typing import List, Optional, Tuple

import numpy as np
from PIL import Image as Pillow
from PIL.Image import Image
from wsidicom import ImageData
from wsidicom.codec import Encoder, LossyCompressionIsoStandard
from wsidicom.geometry import PointMm, Size
from wsidicom.metadata import ImageCoordinateSystem


class DicomizerImageData(ImageData, metaclass=ABCMeta):
    """
    Metaclass for Dicomizer image data. Subclasses should implement all the abstract
    methods and properties in the base ImageData-class, and the pyramid_index property.
    """

    _default_z = None

    @property
    @abstractmethod
    def pyramid_index(self) -> int:
        """Should return pyramid level for image data."""
        raise NotImplementedError()

    @property
    def image_coordinate_system(self) -> ImageCoordinateSystem:
        """Return a default ImageCoordinateSystem."""
        return ImageCoordinateSystem(PointMm(0, 0), 90)

    @property
    def bits(self) -> int:
        return self.encoder.bits

    @property
    def lossy_compression(
        self,
    ) -> Optional[List[Tuple[LossyCompressionIsoStandard, float]]]:
        """Return None as image compression is for most format not known."""
        return None

    @property
    def transcoder(self) -> Optional[Encoder]:
        """Return encoder as for most format transcoding is used."""
        return self.encoder

    def _encode(self, image_data: np.ndarray) -> bytes:
        """Return image data encoded in jpeg using set quality and subsample
        options.

        Parameters
        ----------
        image_data: np.ndarray
            Image data to encode.

        Returns
        ----------
        bytes
            Jpeg bytes.
        """
        return self.encoder.encode(image_data)

    def _get_blank_encoded_frame(self, size: Size) -> bytes:
        """Return cached blank encoded frame for size, or create frame if
        cached frame not available or of wrong size.

        Parameters
        ----------
        size: Size
            Size of frame to get.

        Returns
        ----------
        bytes
            Encoded blank frame.
        """
        if self._blank_encoded_frame_size != size:
            frame = np.full(
                size.to_tuple() + (3,), self.blank_color, dtype=np.dtype(np.uint8)
            )
            self._blank_encoded_frame = self.encoder.encode(frame)
            self._blank_encoded_frame_size = size
        return self._blank_encoded_frame

    def _get_blank_decoded_frame(self, size: Size) -> Image:
        """Return cached blank decoded frame for size, or create frame if
        cached frame not available or of wrong size.

        Parameters
        ----------
        size: Size
            Size of frame to get.

        Returns
        ----------
        bytes
            Decoded blank frame.
        """
        if self._blank_decoded_frame is None or self._blank_decoded_frame_size != size:
            frame = Pillow.new("RGB", size.to_tuple(), self.blank_color)
            self._blank_decoded_frame = frame
            self._blank_decoded_frame_size = size
        return self._blank_decoded_frame

    def _detect_blank_tile(self, tile: np.ndarray) -> bool:
        """Detect if tile is a blank tile, i.e. is filled with background color.
        First checks if the corners  before checking whole tile.

        Parameters
        ----------
        tile: np.ndarray
            Tile to check if blank.

        Returns
        ----------
        bool
            True if tile is blank.
        """

        TOP = RIGHT = -1
        BOTTOM = LEFT = 0
        CORNERS_Y = [BOTTOM, BOTTOM, TOP, TOP]
        CORNERS_X = [LEFT, RIGHT, LEFT, RIGHT]
        background = np.array(self.blank_color)
        corners_rgb = np.ix_(CORNERS_X, CORNERS_Y)
        if np.all(tile[corners_rgb] == background):
            if np.all(tile == background):
                return True
        return False



#    Copyright 2024 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Image data for pyisintax compatible file."""

from io import BytesIO
from typing import List, Optional, Tuple, Union

import numpy as np
from PIL import Image as Pillow
from PIL.Image import Image
from pydicom.uid import UID, JPEGBaseline8Bit
from wsidicom.codec import Encoder
from wsidicom.errors import WsiDicomNotFoundError
from wsidicom.geometry import Point, PointMm, Region, Size, SizeMm
from wsidicom.metadata import Image as ImageMetadata
from wsidicom.metadata import ImageCoordinateSystem

from isyntax import ISyntax
from wsidicomizer.image_data import DicomizerImageData


class ISyntaxLevelImageData(DicomizerImageData):
    def __init__(
        self,
        isyntax: ISyntax,
        merged_metadata: ImageMetadata,
        tile_size: Optional[int],
        encoder: Encoder,
        level: int,
    ):
        super().__init__(encoder)
        self._slide = isyntax
        self._slide_level = isyntax.wsi.get_level(level)
        pixel_spacing = SizeMm(self._slide_level.mpp_x, self._slide_level.mpp_y) / 1000
        if (
            merged_metadata.pixel_spacing is not None
            and merged_metadata.pixel_spacing != pixel_spacing
        ):
            # Override pixel spacing
            self._pixel_spacing = merged_metadata.pixel_spacing
        else:
            self._pixel_spacing = pixel_spacing
        self._file_tile_size = Size(self._slide.tile_width, self._slide.tile_height)
        if tile_size is None:
            self._tile_size = self._file_tile_size
        else:
            self._tile_size = Size(tile_size, tile_size)
        self._level = level
        self._blank_encoded_frame = bytes()
        self._blank_encoded_frame_size = None
        self._blank_decoded_frame = None
        self._blank_decoded_frame_size = None

    @property
    def image_size(self) -> Size:
        return Size(self._slide_level.width, self._slide_level.height)

    @property
    def tile_size(self) -> Size:
        return self._tile_size

    @property
    def file_tile_size(self) -> Size:
        return self._file_tile_size

    @property
    def pixel_spacing(self) -> SizeMm:
        return self._pixel_spacing

    @property
    def downsample(self) -> float:
        return self._slide_level.scale

    @property
    def pyramid_index(self) -> int:
        return self._level

    @property
    def transfer_syntax(self) -> UID:
        """The uid of the transfer syntax of the image."""
        return self.encoder.transfer_syntax

    @property
    def photometric_interpretation(self) -> str:
        return self.encoder.photometric_interpretation

    @property
    def samples_per_pixel(self) -> int:
        return self.encoder.samples_per_pixel

    @property
    def focal_planes(self) -> List[float]:
        return [0.0]

    @property
    def optical_paths(self) -> List[str]:
        return ["0"]

    @property
    def blank_color(self) -> Union[int, Tuple[int, int, int]]:
        return (255, 255, 255)

    @property
    def image_coordinate_system(self) -> ImageCoordinateSystem:
        return ImageCoordinateSystem(PointMm(0, 0), 0)

    @property
    def thread_safe(selt) -> bool:
        return False

    def stitch_tiles(self, region: Region, path: str, z: float, threads: int) -> Image:
        """Overrides ImageData stitch_tiles() to read reagion directly from
        ISyntax object.

        Parameters
        ----------
        region: Region
             Pixel region to stitch to image
        path: str
            Optical path
        z: float
            Z coordinate
        threads: int
            Threads to use for stiching, not used in this implementation.

        Returns
        ----------
        Image
            Stitched image
        """
        if z not in self.focal_planes:
            raise WsiDicomNotFoundError(f"focal plane {z}", str(self))
        if path not in self.optical_paths:
            raise WsiDicomNotFoundError(f"optical path {path}", str(self))
        image_data = self._get_region(region)
        if image_data is None:
            return self._get_blank_decoded_frame(region.size)
        return Pillow.fromarray(image_data)

    def _get_region(self, region: Region) -> Optional[np.ndarray]:
        """Return Image read from region in ISyntax image. If image data for
        region is blank, None is returned.

        Parameters
        ----------
        region: Region
            Region to get image for.

        Returns
        ----------
        Optional[np.ndarray]
            Image data of region, or None if region is blank.
        """
        if region.size.width < 0 or region.size.height < 0:
            raise ValueError("Negative size not allowed")

        region_data = self._slide.read_region(
            region.start.x,
            region.start.y,
            region.size.width,
            region.size.height,
            self._level,
        )[:, :, :3]
        if self._detect_blank_tile(region_data):
            return None
        return region_data

    def _get_tile(self, tile_point: Point, z: float, path: str) -> Optional[np.ndarray]:
        if z not in self.focal_planes:
            raise WsiDicomNotFoundError(f"focal plane {z}", str(self))
        if path not in self.optical_paths:
            raise WsiDicomNotFoundError(f"optical path {path}", str(self))
        if self._tile_size == self.file_tile_size:
            tile = self._slide.read_tile(tile_point.x, tile_point.y, self._level)[
                :, :, :3
            ]
        else:
            tile = self._slide.read_region(
                tile_point.x * self._tile_size.width,
                tile_point.y * self._tile_size.height,
                self._tile_size.width,
                self._tile_size.height,
                self._level,
            )[:, :, :3]
        if self._detect_blank_tile(tile):
            return None
        return tile

    def _get_encoded_tile(self, tile_point: Point, z: float, path: str) -> bytes:
        """Return image bytes for tile.

        Parameters
        ----------
        tile_point: Point
            Tile position to get.
        z: float
            Focal plane of tile to get.
        path: str
            Optical path of tile to get.

        Returns
        ----------
        bytes
            Tile bytes.
        """
        tile = self._get_tile(tile_point, z, path)
        if tile is None:
            return self._get_blank_encoded_frame(self.tile_size)
        return self.encoder.encode(tile)

    def _get_decoded_tile(self, tile_point: Point, z: float, path: str) -> Image:
        """Return Image for tile.

        Parameters
        ----------
        tile_point: Point
            Tile position to get.
        z: float
            Focal plane of tile to get.
        path: str
            Optical path of tile to get.

        Returns
        ----------
        Image
            Tile as Image.
        """
        tile = self._get_tile(tile_point, z, path)
        if tile is None:
            return self._get_blank_decoded_frame(self.tile_size)
        return Pillow.fromarray(tile)


class ISyntaxAssociatedImageImageData(DicomizerImageData):
    def __init__(
        self,
        frame: bytes,
        encoder: Encoder,
        force_transcoding: bool = False,
    ):
        super().__init__(encoder)
        self._frame = frame
        self._force_transcoding = force_transcoding
        if self._force_transcoding:
            self._transfer_syntax = self._encoder.transfer_syntax
        else:
            self._transfer_syntax = JPEGBaseline8Bit

    @property
    def transfer_syntax(self) -> UID:
        if self._force_transcoding:
            return self._encoder.transfer_syntax
        return JPEGBaseline8Bit

    @property
    def photometric_interpretation(self) -> str:
        if self._force_transcoding:
            return self.encoder.photometric_interpretation
        return "YBR_FULL_422"

    @property
    def image(self) -> Image:
        return Pillow.open(BytesIO(self._frame))

    @property
    def image_size(self) -> Size:
        return Size(self.image.width, self.image.height)

    @property
    def tile_size(self) -> Size:
        return self.image_size

    @property
    def pixel_spacing(self) -> Optional[SizeMm]:
        return None

    @property
    def samples_per_pixel(self) -> int:
        return 3

    @property
    def focal_planes(self) -> List[float]:
        return [0.0]

    @property
    def optical_paths(self) -> List[str]:
        return ["0"]

    @property
    def pyramid_index(self) -> int:
        return 0

    @property
    def image_coordinate_system(self) -> ImageCoordinateSystem:
        return ImageCoordinateSystem(PointMm(0, 0), 0)

    @property
    def thread_safe(self) -> bool:
        return True

    def _get_encoded_tile(self, tile: Point, z: float, path: str) -> bytes:
        if z not in self.focal_planes or path not in self.optical_paths:
            raise WsiDicomNotFoundError(
                f"focal plane {z} or optical path {path}", str(self)
            )
        if self._force_transcoding:
            return self.encoder.encode(self.image)
        return self._frame

    def _get_decoded_tile(self, tile: Point, z: float, path: str) -> Image:
        return self.image



#    Copyright 2022 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module containing settings for WsiDicomizer."""


class Settings:
    """Class containing settings. Settings are to be accessed through the
    global variable settings."""

    def __init__(self) -> None:
        self._default_tile_size = 512
        self._czi_block_cache_size = 8
        self._insert_icc_profile_if_missing = True

    @property
    def default_tile_size(self) -> int:
        """Default tile size to use."""
        return self._default_tile_size

    @default_tile_size.setter
    def default_tile_size(self, value: int) -> None:
        self._default_tile_size = value

    @property
    def czi_block_cache_size(self) -> int:
        """Size of block cache to use for czi files."""
        return self._czi_block_cache_size

    @czi_block_cache_size.setter
    def czi_block_cache_size(self, value: int) -> None:
        self._czi_block_cache_size = value

    @property
    def insert_icc_profile_if_missing(self) -> bool:
        """Whether to insert a default ICC profile in the DICOM file if no profile
        is present in the source file or provided metadata."""
        return self._insert_icc_profile_if_missing

    @insert_icc_profile_if_missing.setter
    def insert_icc_profile_if_missing(self, value: bool) -> None:
        self._insert_icc_profile_if_missing = value


settings = Settings()
"""Global settings variable."""



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Metadata for opentile file."""

from typing import Optional

from opentile import Metadata
from wsidicom.geometry import PointMm
from wsidicom.metadata import Equipment, Image, ImageCoordinateSystem, OpticalPath

from wsidicomizer.metadata import WsiDicomizerMetadata


class OpenTileMetadata(WsiDicomizerMetadata):
    def __init__(self, metadata: Metadata, icc_profile: Optional[bytes] = None):
        equipment = Equipment(
            metadata.scanner_manufacturer,
            metadata.scanner_model,
            metadata.scanner_serial_number,
            metadata.scanner_software_versions,
        )
        image_coordinate_system = None
        if metadata.image_offset is not None:
            image_coordinate_system = ImageCoordinateSystem(
                origin=PointMm(metadata.image_offset[0], metadata.image_offset[1]),
                rotation=0,
            )
        image = Image(
            metadata.aquisition_datetime,
            image_coordinate_system=image_coordinate_system,
        )
        if icc_profile is not None:
            optical_path = OpticalPath(icc_profile=icc_profile)
            optical_paths = [optical_path]
        else:
            optical_paths = None
        super().__init__(equipment=equipment, image=image, optical_paths=optical_paths)



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Metadata for openslide file."""

from typing import cast

from tiffslide import TiffSlide

from wsidicomizer.extras.openslide.openslide import (
    PROPERTY_NAME_BOUNDS_X,
    PROPERTY_NAME_BOUNDS_Y,
    PROPERTY_NAME_MPP_X,
    PROPERTY_NAME_MPP_Y,
    PROPERTY_NAME_OBJECTIVE_POWER,
    PROPERTY_NAME_VENDOR,
    OpenSlide,
)
from wsidicomizer.sources.tiffslide.tiffslide_metadata import OpenSlideLikeMetadata


class OpenSlideMetadata(OpenSlideLikeMetadata):
    def __init__(self, slide: OpenSlide):
        super().__init__(cast(TiffSlide, slide))

    @property
    def bounds_x_property_name(self) -> str:
        return PROPERTY_NAME_BOUNDS_X

    @property
    def bounds_y_property_name(self) -> str:
        return PROPERTY_NAME_BOUNDS_Y

    @property
    def mpp_x_property_name(self) -> str:
        return PROPERTY_NAME_MPP_X

    @property
    def mpp_y_property_name(self) -> str:
        return PROPERTY_NAME_MPP_Y

    @property
    def objective_power_property_name(self) -> str:
        return PROPERTY_NAME_OBJECTIVE_POWER

    @property
    def vendor_property_name(self) -> str:
        return PROPERTY_NAME_VENDOR



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Image data for czi file."""

from collections import defaultdict
from dataclasses import dataclass
from functools import cached_property, lru_cache
from pathlib import Path
from threading import RLock
from typing import Dict, List, Optional, Tuple

import numpy as np
from czifile import CziFile, DirectoryEntryDV
from PIL import Image as Pillow
from PIL.Image import Image
from pydicom.uid import UID
from wsidicom.codec import Encoder
from wsidicom.geometry import Point, Region, Size, SizeMm
from wsidicom.metadata import Image as ImageMetadata

from wsidicomizer.config import settings
from wsidicomizer.image_data import DicomizerImageData
from wsidicomizer.sources.czi.czi_metadata import CziMetadata


@dataclass(frozen=True)
class CziBlock:
    index: int
    start: Point
    size: Size


class CziImageData(DicomizerImageData):
    def __init__(
        self,
        czi: CziFile,
        tile_size: Optional[int],
        encoder: Encoder,
        czi_metadata: CziMetadata,
        merged_metadata: ImageMetadata,
    ) -> None:
        """Wraps a czi file to ImageData. Multiple pyramid levels are currently
        not supported.

        Parameters
        ----------
        filepath: str
            Path to czi file to wrap.
        tile_size: int
            Output tile size.
        encoded: Encoder
            Encoded to use.
        czi_metadata: CziMetadata
            Czi metadata to use.
        merged_metadata: ImageMetadata
            Merged image metadata to use.
        """
        self._czi = czi
        self._czi_metadata = czi_metadata
        self._merged_metadata = merged_metadata

        assert self._merged_metadata.pixel_spacing is not None
        self._czi._fh.lock = True
        self._dtype = self._czi.dtype
        super().__init__(encoder)
        if tile_size is None:
            tile_size = settings.default_tile_size
        self._tile_size = Size(tile_size, tile_size)
        assert isinstance(self._czi.filtered_subblock_directory, list)
        self._block_directory = self._czi.filtered_subblock_directory
        self._block_locks: Dict[int, RLock] = defaultdict(RLock)

        if self._merged_metadata.pixel_spacing is None:
            raise ValueError("Could not determine pixel spacing for czi image.")
        self._pixel_spacing = self._merged_metadata.pixel_spacing
        self._image_coordinate_system = merged_metadata.image_coordinate_system

    @property
    def pyramid_index(self) -> int:
        return 0

    @property
    def transfer_syntax(self) -> UID:
        return self.encoder.transfer_syntax

    @cached_property
    def photometric_interpretation(self) -> str:
        return self.encoder.photometric_interpretation

    @property
    def pixel_spacing(self) -> SizeMm:
        return self._pixel_spacing

    @property
    def thread_safe(self) -> bool:
        return True

    @cached_property
    def image_size(self) -> Size:
        """The pixel size of the image."""
        return Size(self._get_size(axis="X"), self._get_size(axis="Y"))

    @property
    def tile_size(self) -> Size:
        """The pixel tile size of the image."""
        return self._tile_size

    @cached_property
    def tiled_size(self) -> Size:
        return self.image_size.ceil_div(self.tile_size)

    @cached_property
    def focal_planes(self) -> List[float]:
        """Focal planes available in the image defined in um."""
        return sorted(self._czi_metadata.focal_plane_mapping)

    @property
    def optical_paths(self) -> List[str]:
        """Optical paths available in the image."""
        return self._czi_metadata.channel_mapping

    @cached_property
    def blank_decoded_tile(self) -> Image:
        return Pillow.fromarray(self._create_blank_tile())

    @cached_property
    def blank_encoded_tile(self) -> bytes:
        return self.encoder.encode(self._create_blank_tile())

    @cached_property
    def pixel_origin(self) -> Point:
        """Return coordinate of the top-left of the image."""
        return Point(self._get_start(axis="X"), self._get_start(axis="Y"))

    @cached_property
    def tile_directory(self) -> Dict[Tuple[Point, float, str], List[CziBlock]]:
        """Return dict of block, block start, and block size by tile position,
        focal plane and optical path

        Returns
        ----------
        Dict[Tuple[Point, float, str], Sequence[CziBlock]]:
            Directory of tile point, focal plane and channel as key and
            list of block, block start, and block size as item.
        """
        tile_directory: Dict[Tuple[Point, float, str], List[CziBlock]] = defaultdict(
            list
        )
        assert isinstance(self._czi.filtered_subblock_directory, list)
        for index, block in enumerate(self._czi.filtered_subblock_directory):
            block_start, block_size, z, c = self._get_block_dimensions(block)
            tile_region = Region.from_points(
                block_start // self.tile_size,
                (block_start + block_size).ceil_div(self.tile_size),
            )
            for tile in tile_region.iterate_all():
                tile_directory[tile, z, c].append(
                    CziBlock(index, block_start, block_size)
                )

        return tile_directory

    @cached_property
    def samples_per_pixel(self) -> int:
        return self._get_size(axis="0")

    @property
    def block_directory(self) -> List[DirectoryEntryDV]:
        return self._block_directory

    @staticmethod
    def detect_format(filepath: Path) -> Optional[str]:
        try:
            with CziFile(filepath):
                return "czi"
        except ValueError:
            return None

    def _get_tile(self, tile_point: Point, z: float, path: str) -> np.ndarray:
        """Return tile data as numpy array for tile.

        Parameters
        ----------
        tile_point: Point
            Tile coordinate to get.

        Returns
        ----------
        np.ndarray
            Tile as numpy array.
        """
        # A blank tile to paste blocks into
        image_data = self._create_blank_tile()
        if (tile_point, z, path) not in self.tile_directory:
            # Should not happen (get_decoded_tile() and get_enoded_tile()
            # should already have checked).
            return image_data

        # For each block covering the tile
        for block in self.tile_directory[tile_point, z, path]:
            # Start and end coordinates for block and tile
            block_end = block.start + block.size
            tile_start = tile_point * self.tile_size
            tile_end = (tile_point + 1) * self.tile_size

            # The block and tile both cover the region between these points
            tile_block_start_intersection = Point.max(tile_start, block.start)
            tile_block_end_intersection = Point.min(tile_end, block_end)

            # The intersects in relation to block and tile origin
            block_start_in_tile = tile_block_start_intersection - tile_start
            block_end_in_tile = tile_block_end_intersection - tile_start
            tile_start_in_block = tile_block_start_intersection - block.start
            tile_end_in_block = tile_block_end_intersection - block.start

            # Get decompressed data
            block_data = self._get_tile_data(block.index)
            # Reshape the block data to remove leading 1-indices.
            block_data.shape = self._size_to_numpy_shape(block.size)
            # Paste in block data into tile.
            image_data[
                block_start_in_tile.y : block_end_in_tile.y,
                block_start_in_tile.x : block_end_in_tile.x,
            ] = block_data[
                tile_start_in_block.y : tile_end_in_block.y,
                tile_start_in_block.x : tile_end_in_block.x,
            ]
        return image_data

    def _get_decoded_tile(self, tile: Point, z: float, path: str) -> Image:
        """Return Image for tile.

        Parameters
        ----------
        tile: Point
            Tile position to get.
        z: float
            Focal plane of tile to get.
        path: str
            Optical path of tile to get.

        Returns
        ----------
        Image
            Tile as Image.
        """
        if (tile, z, path) not in self.tile_directory:
            return self.blank_decoded_tile
        return Pillow.fromarray(self._get_tile(tile, z, path))

    def _get_encoded_tile(self, tile: Point, z: float, path: str) -> bytes:
        """Return image bytes for tile. Tile is encoded as jpeg.

        Parameters
        ----------
        tile: Point
            Tile position to get.
        z: float
            Focal plane of tile to get.
        path: str
            Optical path of tile to get.

        Returns
        ----------
        bytes
            Tile bytes.
        """
        if (tile, z, path) not in self.tile_directory:
            return self.blank_encoded_tile
        frame = self._get_tile(tile, z, path)
        return self.encoder.encode(frame)

    def _get_size(self, axis: str) -> int:
        index = self._get_axis_index(axis)
        assert isinstance(self._czi.shape, tuple)
        return self._czi.shape[index]

    def _get_start(self, axis: str) -> int:
        index = self._get_axis_index(axis)
        assert isinstance(self._czi.start, tuple)
        return self._czi.start[index]

    def _get_axis_index(self, axis: str) -> int:
        index = str(self._czi.axes).index(axis.capitalize())
        if not index >= 0:
            raise ValueError(f"Axis {axis} not found in axes {self._czi.axes}")
        return index

    def _create_blank_tile(self) -> np.ndarray:
        """Return blank tile in numpy array.

        Returns
        ----------
        np.ndarray
            A blank tile as numpy array.
        """
        if self.photometric_interpretation == "MONOCHROME2":
            fill_value = 0
        else:
            fill_value = 1
        assert isinstance(self._czi.dtype, np.dtype)
        return np.full(
            self._size_to_numpy_shape(self.tile_size),
            fill_value * np.iinfo(self._czi.dtype).max,
            dtype=np.dtype(self._czi.dtype),
        )

    @lru_cache(settings.czi_block_cache_size)
    def _get_tile_data(self, block_index: int) -> np.ndarray:
        """Get decompressed tile data from czi file. Cache the tile data. To
        prevent multiple threads proceseing the same tile, use a lock for
        each block."""
        block_lock = self._block_locks[block_index]
        try:
            # Try to lock block.
            aquired = block_lock.acquire(blocking=False)
            if not aquired:
                # Another thread is already reading the block.
                # Wait for lock and hopefully read from cache.
                aquired = block_lock.acquire(blocking=True)
                return self._get_tile_data(block_index)
            else:
                # Read the block data.
                block = self.block_directory[block_index]
                return block.data_segment().data()
        finally:
            if aquired:
                block_lock.release()

    def _get_block_dimensions(
        self, block: DirectoryEntryDV
    ) -> Tuple[Point, Size, float, str]:
        """Return start coordinate and size for block relative to image
        origin and block focal plane and optical path.

        Parameters
        ----------
        block: DirectoryEntryDV
            Block to get start and size from.

        Returns
        ----------
        Tuple[Point, Size]
            Start point coordinate, size, focal plane and optical path for
            block.
        """
        x_start: Optional[int] = None
        x_size: Optional[int] = None
        y_start: Optional[int] = None
        y_size: Optional[int] = None
        z = 0.0
        c = "0"

        for dimension_entry in block.dimension_entries:
            if dimension_entry.dimension == "X":
                x_start = dimension_entry.start
                x_size = dimension_entry.size
            elif dimension_entry.dimension == "Y":
                y_start = dimension_entry.start
                y_size = dimension_entry.size
            elif dimension_entry.dimension == "Z":
                z = self._czi_metadata.focal_plane_mapping[dimension_entry.start]
            elif dimension_entry.dimension == "C":
                c = self._czi_metadata.channel_mapping[dimension_entry.start]

        if x_start is None or x_size is None or y_start is None or y_size is None:
            raise ValueError("Could not determine position of block.")

        return (Point(x_start, y_start) - self.pixel_origin, Size(x_size, y_size), z, c)

    def _size_to_numpy_shape(self, size: Size) -> Tuple[int, ...]:
        """Return a tuple for use with numpy.shape."""
        if self.samples_per_pixel == 1:
            return size.height, size.width
        return size.height, size.width, self.samples_per_pixel



#    Copyright 2024 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Source for reading libisyntax compatible file."""

from pathlib import Path
from typing import Dict, Optional

from wsidicom.codec import Encoder
from wsidicom.metadata import WsiMetadata

from isyntax import ISyntax
from wsidicomizer.dicomizer_source import DicomizerSource
from wsidicomizer.extras.isyntax.isyntax_image_data import (
    ISyntaxAssociatedImageImageData,
    ISyntaxLevelImageData,
)
from wsidicomizer.extras.isyntax.isyntax_metadata import ISyntaxMetadata
from wsidicomizer.image_data import DicomizerImageData
from wsidicomizer.metadata import WsiDicomizerMetadata


class ISyntaxSource(DicomizerSource):
    def __init__(
        self,
        filepath: Path,
        encoder: Encoder,
        tile_size: Optional[int] = None,
        metadata: Optional[WsiMetadata] = None,
        default_metadata: Optional[WsiMetadata] = None,
        include_confidential: bool = True,
        force_transcoding: bool = False,
        cache: int = 2048,
    ) -> None:
        """Create a new ISyntaxSource.

        Parameters
        ----------
        filepath: Path
            Path to the file.
        encoder: Encoder
            Encoder to use. Pyramid is always re-encoded using the encoder."""
        self._slide = ISyntax.open(filepath, cache)
        self._force_transcoding = force_transcoding
        self._base_metadata = ISyntaxMetadata(self._slide)
        super().__init__(
            filepath,
            encoder,
            tile_size,
            metadata,
            default_metadata,
            include_confidential,
        )

    @staticmethod
    def is_supported(path: Path) -> bool:
        try:
            ISyntax.open(path)
        except Exception:
            return False
        return True

    @property
    def base_metadata(self) -> WsiDicomizerMetadata:
        return self._base_metadata

    @property
    def pyramid_levels(self) -> Dict[int, int]:
        return {0: 0}

    def _create_level_image_data(self, level_index: int) -> DicomizerImageData:
        return ISyntaxLevelImageData(
            self._slide,
            self.metadata.image,
            self._tile_size,
            self._encoder,
            level_index,
        )

    def _create_label_image_data(self) -> Optional[DicomizerImageData]:
        label = self._slide.read_label_image_jpeg()
        if label is None:
            return None
        return ISyntaxAssociatedImageImageData(
            label.tobytes(), self._encoder, self._force_transcoding
        )

    def _create_overview_image_data(self) -> Optional[DicomizerImageData]:
        overview = self._slide.read_macro_image_jpeg()
        if overview is None:
            return None
        return ISyntaxAssociatedImageImageData(
            overview.tobytes(), self._encoder, self._force_transcoding
        )

    def _create_thumbnail_image_data(self) -> Optional[DicomizerImageData]:
        return None

    def close(self) -> None:
        return self._slide.close()



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Metadata for tiffslide file."""

import logging
from abc import abstractmethod

from tiffslide import TiffSlide
from tiffslide.tiffslide import (
    PROPERTY_NAME_BOUNDS_X,
    PROPERTY_NAME_BOUNDS_Y,
    PROPERTY_NAME_MPP_X,
    PROPERTY_NAME_MPP_Y,
    PROPERTY_NAME_OBJECTIVE_POWER,
    PROPERTY_NAME_VENDOR,
)
from wsidicom.geometry import PointMm, SizeMm
from wsidicom.metadata import (
    Equipment,
    Image,
    ImageCoordinateSystem,
    Objectives,
    OpticalPath,
)

from wsidicomizer.metadata import WsiDicomizerMetadata


class OpenSlideLikeMetadata(WsiDicomizerMetadata):
    @property
    @abstractmethod
    def bounds_x_property_name(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def bounds_y_property_name(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def mpp_x_property_name(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def mpp_y_property_name(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def objective_power_property_name(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def vendor_property_name(self) -> str:
        raise NotImplementedError()

    def __init__(self, slide: TiffSlide):
        magnification = slide.properties.get(self.objective_power_property_name)
        if magnification is not None:
            OpticalPath("0", objective=Objectives(objective_power=float(magnification)))
        equipment = Equipment(
            manufacturer=slide.properties.get(self.vendor_property_name)
        )
        try:
            base_mpp_x = float(slide.properties[self.mpp_x_property_name])
            base_mpp_y = float(slide.properties[self.mpp_y_property_name])
            pixel_spacing = SizeMm(
                base_mpp_x / 1000.0,
                base_mpp_y / 1000.0,
            )
        except (KeyError, TypeError):
            logging.warning(
                f"Could not determine pixel spacing as {slide} did not "
                "provide mpp from the file.",
                exc_info=True,
            )
            pixel_spacing = None
        # Get set image origin and size to bounds if available
        bounds_x = slide.properties.get(self.bounds_x_property_name, None)
        bounds_y = slide.properties.get(self.bounds_y_property_name, None)
        if bounds_x is not None and bounds_y is not None and pixel_spacing is not None:
            origin = PointMm(
                int(bounds_x) * pixel_spacing.width,
                int(bounds_y) * pixel_spacing.height,
            )
            image_coordinate_system = ImageCoordinateSystem(
                origin,
                0,
            )
        else:
            image_coordinate_system = None
        image = Image(
            pixel_spacing=pixel_spacing, image_coordinate_system=image_coordinate_system
        )
        if slide.color_profile is not None:
            optical_path = OpticalPath(icc_profile=slide.color_profile.tobytes())
            optical_paths = [optical_path]
        else:
            optical_paths = None
        super().__init__(equipment=equipment, image=image, optical_paths=optical_paths)


class TiffSlideMetadata(OpenSlideLikeMetadata):
    @property
    def bounds_x_property_name(self) -> str:
        return PROPERTY_NAME_BOUNDS_X

    @property
    def bounds_y_property_name(self) -> str:
        return PROPERTY_NAME_BOUNDS_Y

    @property
    def mpp_x_property_name(self) -> str:
        return PROPERTY_NAME_MPP_X

    @property
    def mpp_y_property_name(self) -> str:
        return PROPERTY_NAME_MPP_Y

    @property
    def objective_power_property_name(self) -> str:
        return PROPERTY_NAME_OBJECTIVE_POWER

    @property
    def vendor_property_name(self) -> str:
        return PROPERTY_NAME_VENDOR



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Source for reading openslide compatible file."""

import math
from pathlib import Path
from typing import Dict, Optional

from wsidicom.codec import Encoder
from wsidicom.metadata.wsi import WsiMetadata

from wsidicomizer.dicomizer_source import DicomizerSource
from wsidicomizer.extras.openslide.openslide import OpenSlide
from wsidicomizer.extras.openslide.openslide_image_data import (
    OpenSlideAssociatedImageData,
    OpenSlideAssociatedImageType,
    OpenSlideLevelImageData,
    OpenSlideThumbnailImageData,
)
from wsidicomizer.extras.openslide.openslide_metadata import OpenSlideMetadata
from wsidicomizer.image_data import DicomizerImageData


class OpenSlideSource(DicomizerSource):
    def __init__(
        self,
        filepath: Path,
        encoder: Encoder,
        tile_size: Optional[int] = None,
        metadata: Optional[WsiMetadata] = None,
        default_metadata: Optional[WsiMetadata] = None,
        include_confidential: bool = True,
    ) -> None:
        self._slide = OpenSlide(filepath)

        self._pyramid_levels = self._get_pyramid_levels(self._slide)
        self._base_metadata = OpenSlideMetadata(self._slide)
        super().__init__(
            filepath,
            encoder,
            tile_size,
            metadata,
            default_metadata,
            include_confidential,
        )

    def close(self) -> None:
        return self._slide.close()

    @property
    def base_metadata(self) -> WsiMetadata:
        return self._base_metadata

    @property
    def pyramid_levels(self) -> Dict[int, int]:
        return self._pyramid_levels

    @staticmethod
    def is_supported(filepath: Path) -> bool:
        """Return True if file in filepath is supported by OpenSlide."""
        return OpenSlide.detect_format(str(filepath)) is not None

    def _create_level_image_data(self, level_index: int) -> DicomizerImageData:
        return OpenSlideLevelImageData(
            self._slide,
            self.metadata.image,
            level_index,
            self._tile_size,
            self._encoder,
        )

    def _create_label_image_data(self) -> Optional[DicomizerImageData]:
        if (
            OpenSlideAssociatedImageType.LABEL.value
            not in self._slide.associated_images
        ):
            return None
        return OpenSlideAssociatedImageData(
            self._slide, OpenSlideAssociatedImageType.LABEL, self._encoder
        )

    def _create_overview_image_data(self) -> Optional[DicomizerImageData]:
        if (
            OpenSlideAssociatedImageType.MACRO.value
            not in self._slide.associated_images
        ):
            return None
        return OpenSlideAssociatedImageData(
            self._slide, OpenSlideAssociatedImageType.MACRO, self._encoder
        )

    def _create_thumbnail_image_data(self) -> Optional[DicomizerImageData]:
        if (
            OpenSlideAssociatedImageType.THUMBNAIL.value
            not in self._slide.associated_images
        ):
            return None
        return OpenSlideThumbnailImageData(
            self._slide, self.metadata.image, self._encoder
        )

    @staticmethod
    def _get_pyramid_levels(slide: OpenSlide) -> Dict[int, int]:
        """Return list of pyramid levels present in openslide slide."""
        return {
            int(math.log2(int(downsample))): index
            for index, downsample in enumerate(slide.level_downsamples)
        }
